// Copyright (c) 2020 vesoft inc. All rights reserved.
//
// This source code is licensed under Apache 2.0 License.

#include "graph/executor/query/GetVerticesExecutor.h"

using nebula::storage::StorageClient;
using nebula::storage::StorageRpcResponse;
using nebula::storage::cpp2::GetPropResponse;

namespace nebula {
namespace graph {

folly::Future<Status> GetVerticesExecutor::execute() {
  return getVertices();
}

folly::Future<Status> GetVerticesExecutor::getVertices() {
  SCOPED_TIMER(&execTime_);

  auto *gv = asNode<GetVertices>(node());
  StorageClient *storageClient = qctx()->getStorageClient();

  DataSet vertices = buildRequestDataSet(gv);
  if (vertices.rows.empty()) {
    // TODO: add test for empty input.
    return finish(
        ResultBuilder().value(Value(DataSet(gv->colNames()))).iter(Iterator::Kind::kProp).build());
  }

  time::Duration getPropsTime;
  StorageClient::CommonRequestParam param(gv->space(),
                                          qctx()->rctx()->session()->id(),
                                          qctx()->plan()->id(),
                                          qctx()->plan()->isProfileEnabled());
  return DCHECK_NOTNULL(storageClient)
      ->getProps(param,
                 std::move(vertices),
                 gv->props(),
                 nullptr,
                 gv->exprs(),
                 gv->dedup(),
                 gv->orderBy(),
                 gv->limit(qctx()),
                 gv->filter())
      .via(runner())
      .ensure([this, getPropsTime]() {
        SCOPED_TIMER(&execTime_);
        otherStats_.emplace("total_rpc", folly::sformat("{}(us)", getPropsTime.elapsedInUSec()));
      })
      .thenValue([this, gv](StorageRpcResponse<GetPropResponse> &&rpcResp) {
        SCOPED_TIMER(&execTime_);
        addStats(rpcResp, otherStats_);
        return handleResp(std::move(rpcResp), gv->colNames());
      });
}

DataSet GetVerticesExecutor::buildRequestDataSet(const GetVertices *gv) {
  if (gv == nullptr) {
    return nebula::DataSet({kVid});
  }
  // Accept Table such as | $a | $b | $c |... as input which one column indicate
  // src
  auto valueIter = ectx_->getResult(gv->inputVar()).iter();
  return buildRequestDataSetByVidType(valueIter.get(), gv->src(), gv->dedup());
}

}  // namespace graph
}  // namespace nebula
